# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
:mod:`dazl.client` package
==========================

The :mod:`dazl.client` module a high-level view of the Ledger API in a friendly way.

It provides:

  * a callback-based API for interacting with events on the read-side of the Ledger API
  * convenient methods for creating arbitrary commands on the write-side of the Ledger API.

For a higher-level, more declarative API, see :mod:`dazl.query`.
"""

from . import config
from ._base_model import (
    CREATE_IF_MISSING,
    EXCEPTION_IF_MISSING,
    NONE_IF_MISSING,
    ExitCode,
    LedgerRun,
)
from ._network_client_impl import _NetworkImpl
from ._party_client_impl import _PartyClientImpl
from .api import (
    AIOPartyClient,
    Network,
    PartyClient,
    SimplePartyClient,
    async_network,
    simple_client,
)
from .bots import Bot, BotCollection, BotEntry
from .runner import run



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Simple command-line handlers.
"""

import logging
from typing import List, Sequence

from .._logging import configure as configure_logger
from ..client.errors import ConfigurationError
from ._base import CliCommand
from .ls import ListAllCommand
from .metadata import PrintMetadataCommand
from .tail import TailCommand
from .upload import UploadCommand
from .version import VersionCommand

COMMANDS = [
    ListAllCommand(),
    PrintMetadataCommand(),
    TailCommand(),
    UploadCommand(),
    VersionCommand(),
]  # type: List[CliCommand]


def main():
    """
    Executes one of the known commands.
    """
    from sys import argv, exit

    exit(_main(argv))


def _main(argv: "Sequence[str]") -> int:
    if len(argv) > 1:
        command = argv[1]
        command_args = argv[2:]

        for cmd in COMMANDS:
            if cmd.name == command:
                try:
                    return run(cmd, command_args)
                except ConfigurationError as error:
                    for reason in error.reasons:
                        print(reason)
                    return -1

        print("Unknown command: " + command)

    print_cmd_help()
    return -2


def run(cmd, args) -> int:
    parser = cmd.parser()
    parsed_args = parser.parse_args(args)

    log_level = getattr(parsed_args, "log_level", logging.WARNING)
    if log_level is None:
        log_level = logging.WARNING

    configure_logger(level=log_level)
    return cmd.execute(parsed_args)


def print_cmd_help():
    print("Valid commands:")
    for cmd in COMMANDS:
        if not cmd.hidden:
            print(cmd.name)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    Awaitable as __Awaitable,
    Iterable as __Stream,
    Union as __Union,
    overload,
)

from grpc import Channel as __Channel
from grpc.aio import Channel as __AsyncChannel

from .party_management_service_pb2 import (
    AllocatePartyRequest,
    AllocatePartyResponse,
    GetParticipantIdRequest,
    GetParticipantIdResponse,
    GetPartiesRequest,
    GetPartiesResponse,
    ListKnownPartiesRequest,
    ListKnownPartiesResponse,
)

__all__ = [
    "PartyManagementServiceStub",
]

class PartyManagementServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> _PartyManagementServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> _PartyManagementServiceStub_Async: ...
    def GetParticipantId(self, __1: GetParticipantIdRequest) -> __Union[GetParticipantIdResponse, __Awaitable[GetParticipantIdResponse]]: ...
    def GetParties(self, __1: GetPartiesRequest) -> __Union[GetPartiesResponse, __Awaitable[GetPartiesResponse]]: ...
    def ListKnownParties(self, __1: ListKnownPartiesRequest) -> __Union[ListKnownPartiesResponse, __Awaitable[ListKnownPartiesResponse]]: ...
    def AllocateParty(self, __1: AllocatePartyRequest) -> __Union[AllocatePartyResponse, __Awaitable[AllocatePartyResponse]]: ...

class _PartyManagementServiceStub(PartyManagementServiceStub):
    def GetParticipantId(self, __1: GetParticipantIdRequest) -> GetParticipantIdResponse: ...
    def GetParties(self, __1: GetPartiesRequest) -> GetPartiesResponse: ...
    def ListKnownParties(self, __1: ListKnownPartiesRequest) -> ListKnownPartiesResponse: ...
    def AllocateParty(self, __1: AllocatePartyRequest) -> AllocatePartyResponse: ...

class _PartyManagementServiceStub_Async(PartyManagementServiceStub):
    def GetParticipantId(self, __1: GetParticipantIdRequest) -> __Awaitable[GetParticipantIdResponse]: ...
    def GetParties(self, __1: GetPartiesRequest) -> __Awaitable[GetPartiesResponse]: ...
    def ListKnownParties(self, __1: ListKnownPartiesRequest) -> __Awaitable[ListKnownPartiesResponse]: ...
    def AllocateParty(self, __1: AllocatePartyRequest) -> __Awaitable[AllocatePartyResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/daml_lf_dev/daml_lf_0.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="com/daml/daml_lf_dev/daml_lf_0.proto",
    package="daml_lf_0",
    syntax="proto3",
    serialized_options=b"\n\024com.daml.daml_lf_dev\252\002\034Com.Daml.Daml_Lf_Dev.DamlLf0",
    serialized_pb=b'\n$com/daml/daml_lf_dev/daml_lf_0.proto\x12\tdaml_lf_0"\t\n\x07PackageB5\n\x14\x63om.daml.daml_lf_dev\xaa\x02\x1c\x43om.Daml.Daml_Lf_Dev.DamlLf0b\x06proto3',
)


_PACKAGE = _descriptor.Descriptor(
    name="Package",
    full_name="daml_lf_0.Package",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=51,
    serialized_end=60,
)

DESCRIPTOR.message_types_by_name["Package"] = _PACKAGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Package = _reflection.GeneratedProtocolMessageType(
    "Package",
    (_message.Message,),
    {
        "DESCRIPTOR": _PACKAGE,
        "__module__": "com.daml.daml_lf_dev.daml_lf_0_pb2"
        # @@protoc_insertion_point(class_scope:daml_lf_0.Package)
    },
)
_sym_db.RegisterMessage(Package)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import warnings

from ..ledger.aio.pkgloader_compat import PackageLoader, SyncPackageService

warnings.warn(
    "dazl.client.pkg_loader is deprecated; use dazl.ledger.aio instead.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = ["SyncPackageService", "PackageLoader"]



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    Awaitable as __Awaitable,
    Iterable as __Stream,
    Union as __Union,
    overload,
)

from grpc import Channel as __Channel
from grpc.aio import Channel as __AsyncChannel

from .transaction_service_pb2 import (
    GetFlatTransactionResponse,
    GetLedgerEndRequest,
    GetLedgerEndResponse,
    GetTransactionByEventIdRequest,
    GetTransactionByIdRequest,
    GetTransactionsRequest,
    GetTransactionsResponse,
    GetTransactionTreesResponse,
)

__all__ = [
    "TransactionServiceStub"
]

# noinspection PyPep8Naming
class TransactionServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> _TransactionServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> _TransactionServiceStub_Async: ...
    def GetTransactions(self, __1: GetTransactionsRequest) -> __Union[__Stream[GetTransactionsResponse], __AsyncStream[GetTransactionsResponse]]: ...
    def GetTransactionTrees(self, __1: GetTransactionsRequest) -> __Union[__Stream[GetTransactionTreesResponse], __AsyncStream[GetTransactionTreesResponse]]: ...
    def GetTransactionByEventId(self, __1: GetTransactionByEventIdRequest) -> __Union[GetTransactionsResponse, __Awaitable[GetTransactionsResponse]]: ...
    def GetTransactionById(self, __1: GetTransactionByIdRequest) -> __Union[GetTransactionsResponse, __Awaitable[GetTransactionsResponse]]: ...
    def GetFlatTransactionByEventId(self, __1: GetTransactionByEventIdRequest) -> __Union[GetFlatTransactionResponse, __Awaitable[GetFlatTransactionResponse]]: ...
    def GetFlatTransactionById(self, __1: GetTransactionByIdRequest) -> __Union[GetFlatTransactionResponse, __Awaitable[GetFlatTransactionResponse]]: ...
    def GetLedgerEnd(self, __1: GetLedgerEndRequest) -> __Union[GetLedgerEndResponse, __Awaitable[GetLedgerEndResponse]]: ...

class _TransactionServiceStub(TransactionServiceStub):
    def GetTransactions(self, __1: GetTransactionsRequest) -> __Stream[GetTransactionsResponse]: ...
    def GetTransactionTrees(self, __1: GetTransactionsRequest) -> __Stream[GetTransactionTreesResponse]: ...
    def GetTransactionByEventId(self, __1: GetTransactionByEventIdRequest) -> GetTransactionsResponse: ...
    def GetTransactionById(self, __1: GetTransactionByIdRequest) -> GetTransactionsResponse: ...
    def GetFlatTransactionByEventId(self, __1: GetTransactionByEventIdRequest) -> GetFlatTransactionResponse: ...
    def GetFlatTransactionById(self, __1: GetTransactionByIdRequest) -> GetFlatTransactionResponse: ...
    def GetLedgerEnd(self, __1: GetLedgerEndRequest) -> GetLedgerEndResponse: ...

class _TransactionServiceStub_Async(TransactionServiceStub):
    def GetTransactions(self, __1: GetTransactionsRequest) -> __AsyncStream[GetTransactionsResponse]: ...
    def GetTransactionTrees(self, __1: GetTransactionsRequest) -> __AsyncStream[GetTransactionTreesResponse]: ...
    def GetTransactionByEventId(self, __1: GetTransactionByEventIdRequest) -> __Awaitable[GetTransactionsResponse]: ...
    def GetTransactionById(self, __1: GetTransactionByIdRequest) -> __Awaitable[GetTransactionsResponse]: ...
    def GetFlatTransactionByEventId(self, __1: GetTransactionByEventIdRequest) -> __Awaitable[GetFlatTransactionResponse]: ...
    def GetFlatTransactionById(self, __1: GetTransactionByIdRequest) -> __Awaitable[GetFlatTransactionResponse]: ...
    def GetLedgerEnd(self, __1: GetLedgerEndRequest) -> __Awaitable[GetLedgerEndResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
import grpc

from . import (
    command_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2,
)


class CommandServiceStub(object):
    """Command Service is able to correlate submitted commands with completion data, identify timeouts, and return contextual
    information with each tracking result. This supports the implementation of stateless clients.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.SubmitAndWait = channel.unary_unary(
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWait",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
        )
        self.SubmitAndWaitForTransactionId = channel.unary_unary(
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWaitForTransactionId",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionIdResponse.FromString,
        )
        self.SubmitAndWaitForTransaction = channel.unary_unary(
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWaitForTransaction",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionResponse.FromString,
        )
        self.SubmitAndWaitForTransactionTree = channel.unary_unary(
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWaitForTransactionTree",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionTreeResponse.FromString,
        )


class CommandServiceServicer(object):
    """Command Service is able to correlate submitted commands with completion data, identify timeouts, and return contextual
    information with each tracking result. This supports the implementation of stateless clients.
    """

    def SubmitAndWait(self, request, context):
        """Submits a single composite command and waits for its result.
        Propagates the gRPC error of failed submissions including Daml interpretation errors.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        - ``RESOURCE_EXHAUSTED``: if the number of in-flight commands reached the maximum (if a limit is configured)
        - ``UNAVAILABLE``: if the participant is not yet ready to submit commands or if the service has been shut down.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SubmitAndWaitForTransactionId(self, request, context):
        """Submits a single composite command, waits for its result, and returns the transaction id.
        Propagates the gRPC error of failed submissions including Daml interpretation errors.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        - ``RESOURCE_EXHAUSTED``: if the number of in-flight commands reached the maximum (if a limit is configured)
        - ``UNAVAILABLE``: if the participant is not yet ready to submit commands or if the service has been shut down.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SubmitAndWaitForTransaction(self, request, context):
        """Submits a single composite command, waits for its result, and returns the transaction.
        Propagates the gRPC error of failed submissions including Daml interpretation errors.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        - ``RESOURCE_EXHAUSTED``: if the number of in-flight commands reached the maximum (if a limit is configured)
        - ``UNAVAILABLE``: if the participant is not yet ready to submit commands or if the service has been shut down.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SubmitAndWaitForTransactionTree(self, request, context):
        """Submits a single composite command, waits for its result, and returns the transaction tree.
        Propagates the gRPC error of failed submissions including Daml interpretation errors.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id
        - ``INVALID_ARGUMENT``: if the payload is malformed or is missing required fields
        - ``RESOURCE_EXHAUSTED``: if the number of in-flight commands reached the maximum (if a limit is configured)
        - ``UNAVAILABLE``: if the participant is not yet ready to submit commands or if the service has been shut down.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_CommandServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "SubmitAndWait": grpc.unary_unary_rpc_method_handler(
            servicer.SubmitAndWait,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
            response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
        ),
        "SubmitAndWaitForTransactionId": grpc.unary_unary_rpc_method_handler(
            servicer.SubmitAndWaitForTransactionId,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
            response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionIdResponse.SerializeToString,
        ),
        "SubmitAndWaitForTransaction": grpc.unary_unary_rpc_method_handler(
            servicer.SubmitAndWaitForTransaction,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
            response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionResponse.SerializeToString,
        ),
        "SubmitAndWaitForTransactionTree": grpc.unary_unary_rpc_method_handler(
            servicer.SubmitAndWaitForTransactionTree,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.FromString,
            response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionTreeResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "com.daml.ledger.api.v1.CommandService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class CommandService(object):
    """Command Service is able to correlate submitted commands with completion data, identify timeouts, and return contextual
    information with each tracking result. This supports the implementation of stateless clients.
    """

    @staticmethod
    def SubmitAndWait(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWait",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SubmitAndWaitForTransactionId(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWaitForTransactionId",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionIdResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SubmitAndWaitForTransaction(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWaitForTransaction",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SubmitAndWaitForTransactionTree(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.CommandService/SubmitAndWaitForTransactionTree",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_command__service__pb2.SubmitAndWaitForTransactionTreeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import sys
from typing import TYPE_CHECKING, AbstractSet, Any, Collection

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable

if TYPE_CHECKING:
    from .daml_lf_1 import (
        Archive,
        DefDataType,
        DefTemplate,
        DefValue,
        Package,
        PackageRef,
        TypeConName,
    )

__all__ = ["PackageProvider", "SymbolLookup"]


@runtime_checkable
class PackageProvider(Protocol):
    """
    Structural interface for a class that can synchronously provide package data.

    If a requested package could not be found, :class:`PackageNotFoundError` should be raised; the
    functions of this protocol should never return ``None``.
    """

    def package(self, ref: "PackageRef") -> "Package":
        """
        Return the :class:`Package` for the specified :class:`PackageRef`.

        :param ref:
            The package ID to look up.
        :return:
            The Package corresponding to this package ID.
        :raises PackageNotFoundError:
            if the package could not be found
        """
        raise NotImplementedError("PackageProvider.package must be implemented")

    def package_ids(self) -> "AbstractSet[PackageRef]":
        """
        Return all package IDs that are known to this :class:`PackageProvider`.

        Implementations are allowed to return larger sets of packages over time, but should
        generally never omit a package ID that was previously returned.

        :return:
            A set of package IDs.
        """
        raise NotImplementedError("PackageProvider.package_ids must be implemented")


@runtime_checkable
class SymbolLookup(Protocol):
    """
    Structural interface for a class that can synchronously provide type/value information.

    If a requested _package_ could not be found, :class:`PackageNotFoundError` is raised.
    If a requested package is known to this :class:`SymbolLookup` but the name is unknown, then
    :class:`NameNotFoundError` is raised. Callers can use this to disambiguate between errors that
    are retryable (missing packages can be fetched) and non-retryable (names within a known package,
    if they do not currently exist, will NEVER exist).
    """

    def archives(self) -> "Collection[Archive]":
        """
        Return the archives that are known to this lookup.
        """
        raise NotImplementedError("SymbolLookup.archives must be implemented")

    def package_ids(self) -> "AbstractSet[PackageRef]":
        """
        Return the package IDs that are known to this lookup.
        """
        raise NotImplementedError("SymbolLookup.package_ids must be implemented")

    def data_type_name(self, ref: "Any") -> "TypeConName":
        """
        Return the :class:`TypeConName` that refers to a :class:`DefDataType` that is known to
        exist in this lookup.

        If this method succeeds, :meth:`data_type` with the returned :class:`TypeConName` as an
        argument should also always succeed.
        """
        raise NotImplementedError("SymbolLookup.data_type_name must be implemented")

    def data_type(self, ref: "Any") -> "DefDataType":
        """
        Return the :class:`DefDataType` for the specified name.
        """
        raise NotImplementedError("SymbolLookup.data_type must be implemented")

    def value(self, ref: "Any") -> "DefValue":
        """
        Return the :class:`DefValue` for the specified name.
        """
        raise NotImplementedError("SymbolLookup.value must be implemented")

    def template_names(self, ref: "Any") -> "Collection[TypeConName]":
        """
        Return all template names that are currently known that are a match for the query. Either
        :class:`PackageRef` or the template name can be `*`.

        Unlike the other methods of this class, this function should return an empty list in the
        case of a match failure; :class:`PackageNotFoundError` or :class:`NameNotFoundError` are
        never thrown.
        """
        raise NotImplementedError("SymbolLookup.template_names must be implemented")

    def template_name(self, ref: "Any") -> "TypeConName":
        """
        Return the :class:`TypeConName` that refers to a :class:`DefTemplate` that is known to
        exist in this lookup.

        If this method succeeds, :meth:`data_type` with the returned :class:`TypeConName` as an
        argument should also always succeed.
        """
        raise NotImplementedError("SymbolLookup.template_name must be implemented")

    def template(self, ref: "Any") -> "DefTemplate":
        """
        Return the :class:`DefTemplate` for the specified name.
        """
        raise NotImplementedError("SymbolLookup.template must be implemented")



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains utilities to help work with ``asyncio``.
"""
from asyncio import (
    AbstractEventLoop,
    AbstractEventLoopPolicy,
    CancelledError,
    Future,
    InvalidStateError,
    Queue as AQueue,
    ensure_future,
    gather,
    get_event_loop,
    new_event_loop,
)
from dataclasses import dataclass
from functools import partial, wraps
from inspect import isawaitable, iscoroutinefunction
from queue import Queue
import sys
import threading
from typing import (
    Any,
    Awaitable,
    Callable,
    Generator,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    cast,
    no_type_check,
)
import warnings

from .. import LOG
from ..prim.datetime import TimeDeltaLike, to_timedelta

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
U = TypeVar("U")

_PENDING = "PENDING"
_CANCELLED = "CANCELLED"
_FINISHED = "FINISHED"


def non_reentrant(async_fn):
    if not iscoroutinefunction(async_fn):
        raise ValueError("expected a coroutine function")

    calls = []

    @wraps(async_fn)
    async def _wrap(*args, **kwargs):
        if not calls:
            calls.append(None)
            try:
                return await async_fn(*args, **kwargs)
            finally:
                calls.pop()
        else:
            raise InvalidStateError("calls to %r cannot be re-entrant", async_fn)

    return _wrap


def isolated_async(async_fn):
    """
    Wrap an async function in a thread with its own event loop.

    This function runs by itself in an event loop created specifically for this object on its own thread.

    This function can be used as a decorator to convert an ``async def`` method to a simple method that blocks while
    executing the method on a background thread.

    :param async_fn:
        The ``async`` function to invoke.
    :return:
        A function that, when invoked, passes its parameters to the underlying function on a background thread in an
        isolated event loop.
    """
    queue = Queue()

    @wraps(async_fn)
    def invoke(*args, **kwargs):
        """
        Invoke the wrapped ``async`` method on a background thread, passing in the specified
        arguments, and returning the value that the wrapped method returns, or raises an Exception
        if the ``async`` function throws.

        :param args: The positional arguments to pass.
        :param kwargs: The keyword arguments to pass.
        :return: The value from the ``async`` function.
        """
        thread = threading.Thread(target=partial(_main, args, kwargs))
        thread.start()
        thread.join()

        result, typ, value, traceback = queue.get()
        if typ is not None:
            # if we have exception information, then an exception was thrown; rethrow it
            raise typ(value).with_traceback(traceback)

        else:
            # we don't have an exception, so assume the result is what we need to pass back
            return result

    def _main(args, kwargs):
        try:
            loop = new_event_loop()
            result = loop.run_until_complete(async_fn(*args, **kwargs))
            loop.close()

            queue.put_nowait((result, None, None, None))
        except:
            typ, value, traceback = sys.exc_info()
            queue.put_nowait((None, typ, value, traceback))

    return invoke


def await_then(awaitable: "Awaitable[T_co]", func: "Callable[[T_co], U]") -> "Awaitable[U]":
    """
    Call a function on the result of an Awaitable, and then return an Awaitable that is resolved
    with that result.

    If the Awaitable is already "done", then the function is invoked immediately with the result
    of that Awaitable and an immediately-completed Future is returned with the result of that
    callback.

    If either the Awaitable or the function throws an exception, that is returned as a failed
    Future. If the Awaitable is cancelled, the returned Future is also cancelled.

    :param awaitable:
    :param func:
    :return:
    """
    try:
        fut = ensure_future(awaitable)
    except TypeError:
        raise TypeError(f"expected a valid awaitable (got {awaitable} instead)")
    if fut.done():
        if fut.cancelled() or fut.exception() is not None:
            return fut  # type: ignore
        else:
            try:
                return completed(func(fut.result()))
            except Exception as ex:
                fut = get_event_loop().create_future()
                fut.set_exception(ex)
                return fut  # type: ignore
    else:
        g = get_event_loop().create_future()

        def _propagate(_):
            if fut.cancelled():
                g.cancel()
            elif fut.exception() is not None:
                g.set_exception(fut.exception())
            else:
                try:
                    g.set_result(func(fut.result()))
                except Exception as ex2:
                    g.set_exception(ex2)

        fut.add_done_callback(_propagate)
        return g  # type: ignore


@dataclass
class FailedInvocation:
    """Marker used within execute_in_loop to propagate exceptions thrown
    by coro_fn through the result queue in a way that can be easily detected
    by the queue's reader.
    """

    ex: Exception


def execute_in_loop(
    loop: AbstractEventLoop,
    coro_fn: "Callable[[], Union[Awaitable[T_co], T_co]]",
    timeout: "Optional[TimeDeltaLike]" = 30.0,
) -> T_co:
    """
    Run a coroutine in a target loop. Exceptions thrown by the coroutine are
    propagated to the caller. Must NOT be called from a coroutine on the same
    loop!

    :param loop:
    :param coro_fn:
    :param timeout: Seconds to wait for the termination of the coroutine.
    :return: The return value from the coroutine.
    """
    from functools import wraps
    from queue import Queue

    q = Queue()  # type: ignore

    def coro_fn_complete(fut):
        LOG.debug("coro_fn_complete: %s", fut)
        if fut.exception() is None:
            q.put_nowait(fut.result())
        else:
            q.put_nowait(FailedInvocation(fut.exception()))

    @wraps(coro_fn)
    def run():
        if iscoroutinefunction(coro_fn):
            fut = ensure_future(coro_fn())
            fut.add_done_callback(lambda _: coro_fn_complete(fut))

        elif isawaitable(coro_fn):
            fut = ensure_future(coro_fn)
            fut.add_done_callback(lambda _: coro_fn_complete(fut))

        elif callable(coro_fn):
            result = coro_fn()
            if isawaitable(result):
                fut = ensure_future(result)
                fut.add_done_callback(lambda _: coro_fn_complete(fut))
            else:
                q.put_nowait(result)
        else:
            raise ValueError("Received an unknown kind of callback")

    loop.call_soon_threadsafe(run)

    timeout_in_seconds = to_timedelta(timeout).total_seconds()
    result = q.get(timeout=timeout_in_seconds)

    if isinstance(result, FailedInvocation):
        raise result.ex

    return result


def completed(value, loop=None) -> Future:
    if loop is None:
        loop = get_event_loop()
    fut = loop.create_future()
    fut.set_result(value)
    return fut


def failed(exception: BaseException, loop=None) -> Future:
    if loop is None:
        loop = get_event_loop()
    fut = loop.create_future()
    fut.set_exception(exception)
    return fut


def propagate(from_: Future, to: Future) -> None:
    """
    Copy the value that ``from_`` is completed with to ``to``, whenever ``from_`` is completed.
    """
    if from_.done():
        copy_result(from_, to)
    else:
        from_.add_done_callback(lambda _: copy_result(from_, to))


def copy_result(from_: Future, to: Future) -> None:
    if not from_.done():
        raise ValueError("from_ must be a completed Future")
    if to.done():
        raise ValueError("to must NOT be a completed Future")

    if from_.cancelled():
        to.cancel()
    else:
        exception = from_.exception()
        if exception is not None:
            to.set_exception(exception)
        else:
            result = from_.result()
            to.set_result(result)


def to_coroutine(callback: Callable[..., Any]) -> Callable[..., Future]:
    """
    Wrap a function or a coroutine, always producing a coroutine.

    In other words, for a function like

    ```
    def double(x):
        return x + x
    ```

    the result of ``to_coroutine(double)`` would be a function that is equivalent to:

    ```
    async def double(x):
        return x + x
    ```

    For functions that are already coroutines, their behavior is unaffected by the wrapper generated
    from this function.

    :param callback:
        The callback to convert to a coroutine.
    :return:
        A coroutine that calls the specified function.
    """
    if not callable(callback):
        raise ValueError("callback must be callable")

    @wraps(callback)
    def invoke_sync(*args, **kwargs) -> Future:
        loop = get_event_loop()

        try:
            result = callback(*args, **kwargs)
            if isinstance(result, Future):
                return result
            else:
                future = loop.create_future()
                future.set_result(result)
                return future
        except Exception as ex:
            future = loop.create_future()
            future.set_exception(ex)
            return future

    if iscoroutinefunction(callback):
        return lambda *args, **kwargs: ensure_future(callback(*args, **kwargs))
    else:
        return invoke_sync


class LongRunningAwaitable:
    """
    An :class:`Awaitable` that "finishes" once all of the futures that have been added to it are
    finished.
    """

    def __init__(self, awaitables: Optional[Iterable[Awaitable[Any]]] = None):
        self._fut = get_event_loop().create_future()
        self._coros = []  # type: List[Awaitable[Any]]
        if awaitables is not None:
            self.extend(awaitables)

    def append(self, *awaitables: Awaitable[Any]) -> None:
        self.extend(awaitables)

    def extend(self, awaitables: Iterable[Awaitable[Any]]):
        for a in awaitables:
            f = ensure_future(a)
            f.add_done_callback(self._future_finished)
            self._coros.append(f)

    def _future_finished(self, fut):
        self._coros.remove(fut)
        if not self._coros:
            self._fut.set_result(None)

    def __await__(self):
        return self._fut.__await__()


class ServiceQueue(Generic[T]):

    # noinspection PyDeprecation
    def __init__(self):
        warnings.warn(
            "ServiceQueue is deprecated; there is no planned replacement",
            DeprecationWarning,
            stacklevel=2,
        )
        self._q = []  # type: Union[List[Tuple[T, Future]], AQueue[Tuple[T, Future]]]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            self._service_fut = safe_create_future()
        self._prev_fut = None

    # noinspection PyDeprecation
    def put(self, value: T) -> Awaitable[None]:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            fut = safe_create_future()
        if isinstance(self._q, list):
            self._q.append((value, fut))
        else:
            self._q.put_nowait((value, fut))
        return fut

    def start(self) -> None:
        """
        Allow open asynchronous iterators to begin reading data from the queue. Calling this method
        more than once has no effect.
        """
        if not self._service_fut.done():
            existing_items = cast(list, self._q)  # type: List[Tuple[T, Future]]
            self._q = AQueue()
            for item in existing_items:
                self._q.put_nowait(item)
            self._service_fut.set_result(None)

    def stop(self):
        if not self._service_fut.done():
            raise RuntimeError("Cannot stop an unstarted ServiceQueue")

        loop = get_event_loop()
        fut = loop.create_future()
        self._q.put_nowait((None, fut))
        return fut

    def abort(self) -> Sequence[T_co]:
        """
        Gracefully terminate the open iterator as quickly as possible and return all remaining
        elements in the queue.
        """

    async def next(self) -> Optional[T_co]:
        if not self._service_fut.done():
            await self._service_fut

        if self._prev_fut is not None:
            self._prev_fut.set_result(None)
            self._prev_fut = None

        q = cast(AQueue, self._q)  # type: AQueue[Tuple[T_co, Future]]
        value, fut = await q.get()
        if value is None:
            fut.set_result(None)
            return None
        else:
            self._prev_fut = fut
            return value

    def __aiter__(self):
        return self

    async def __anext__(self):
        value = await self.next()
        if value is not None:
            return value
        else:
            raise StopAsyncIteration

    def __bool__(self):
        return len(self) > 0

    def __len__(self):
        if isinstance(self._q, list):
            return len(self._q)
        else:
            return self._q.qsize()

    def __repr__(self):
        return repr(self._q)


class ContextFreeFuture(Awaitable[T_co]):
    """
    An awaitable whose loop is defined at the time of either a :meth:`set_result` call or an
    `await`. These futures are more expensive than normal asyncio futures because they are
    thread-safe.
    """

    _result: Optional[T_co] = None
    _exception: Optional[Exception] = None
    _log_traceback = False
    _asyncio_future_blocking = False

    def __init__(self, *, loop: Optional[AbstractEventLoop] = None):
        """
        Initialize the future.

        The optional ``loop`` argument allows explicitly setting the event loop object used by the
        future. Unlike :class:`Future`, If it is not provided, the future does NOT have a loop set,
        and it is assigned at the time that :class:`set_result` or an `await` expression is called,
        or when :class:`set_loop` is explicitly called.

        You may want to provide a loop ahead of time if you know it, and you're really only
        utilizing this class for its thread-safe nature.

        :param loop: An optional loop to assign to the :class:`ContextFreeFuture`.
        """
        if loop is not None and not isinstance(loop, AbstractEventLoop):
            raise ValueError("The provided loop must be a valid event loop")

        from threading import RLock

        self._lock = RLock()
        self._callbacks = []  # type: List[Callable[[ContextFreeFuture[T_co]], None]]
        self._state = _PENDING
        self.__loop = loop  # type: Optional[AbstractEventLoop]
        self._source_traceback = None

    @property
    def _loop(self):
        """
        Return the current loop.

        Because many functions in the asyncio library directly attempt to access the loop of a
        Future for validation purposes, it is necessary for this property to behave in strange ways.
        If this property is accessed in a context where this is a currently running loop, this
        Future will adopt the current running loop as its own, but ONLY if a loop has not yet been
        set. This will usually do the right thing, but in some cases it may be necessary for callers
        to invoke :meth:`set_loop` directly.
        """
        if self.__loop is None:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                # noinspection PyDeprecation
                self.__loop = get_running_loop()
        return self.__loop

    def set_loop(self, loop: AbstractEventLoop) -> None:
        """
        Set the loop affinity for this future.

        It is an error to call this function more than once with different values for the loop.

        :param loop:
        """
        with self._lock:
            if self.__loop is None:
                self.__loop = loop
            elif self.__loop is not loop:
                raise InvalidStateError("This future is already associated with a specific loop.")

    def cancel(self) -> bool:
        """
        Cancel this future.
        :return:
        """
        with self._lock:
            self._log_traceback = False
            if self._state != _PENDING:
                return False

            self._state = _CANCELLED
            self._schedule_callbacks()
            return True

    def set_result(self, result):
        """Mark the future done and set its result.

        If the future is already done when this method is called, raises
        InvalidStateError.
        """
        with self._lock:
            if self._state != _PENDING:
                raise InvalidStateError("{}: {!r}".format(self._state, self))
            self._result = result
            self._state = _FINISHED
            self._schedule_callbacks()

    def set_exception(self, exception):
        """Mark the future done and set an exception.

        If the future is already done when this method is called, raises
        InvalidStateError.
        """
        with self._lock:
            if self._state != _PENDING:
                raise InvalidStateError("{}: {!r}".format(self._state, self))
            if isinstance(exception, type):
                exception = exception()
            if type(exception) is StopIteration:
                raise TypeError(
                    "StopIteration interacts badly with generators "
                    "and cannot be raised into a Future"
                )
            self._exception = exception
            self._state = _FINISHED
            self._schedule_callbacks()
            self._log_traceback = True

    def add_done_callback(
        self, fn: "Callable[[ContextFreeFuture[T_co]], None]", context=None
    ) -> None:
        if context is not None:
            # This used to be a warning, but in newer versions of Python a context
            # is sometimes supplied and nothing seems to particularly go wrong by
            # simply doing nothing with this information.
            pass
        with self._lock:
            if self._loop is not None and self._state != _PENDING:
                self._loop.call_soon_threadsafe(fn, self)
            else:
                self._callbacks.append(fn)

    def _schedule_callbacks(self):
        with self._lock:
            if self.__loop is None:
                self.__loop = get_event_loop()

            # clear out the callbacks list and prevent further additions
            local_callbacks = self._callbacks
            self._callbacks = ()

            for fn in local_callbacks:
                self.__loop.call_soon_threadsafe(fn, self)

    def cancelled(self):
        """Return True if the future was cancelled."""
        with self._lock:
            return self._state == _CANCELLED

    # Don't implement running(); see http://bugs.python.org/issue18699

    def done(self):
        """Return True if the future is done.

        Done means either that a result / exception are available, or that the
        future was cancelled.
        """
        with self._lock:
            return self._state != _PENDING

    def result(self):
        """Return the result this future represents.

        If the future has been cancelled, raises CancelledError.  If the
        future's result isn't yet available, raises InvalidStateError.  If
        the future is done and has an exception set, this exception is raised.
        """
        with self._lock:
            if self._state == _CANCELLED:
                raise CancelledError
            if self._state != _FINISHED:
                raise InvalidStateError("Result is not ready.")
            self._log_traceback = False
            if self._exception is not None:
                raise self._exception
            return self._result

    def exception(self):
        """Return the exception that was set on this future.

        The exception (or None if no exception was set) is returned only if
        the future is done.  If the future has been cancelled, raises
        CancelledError.  If the future isn't done yet, raises
        InvalidStateError.
        """
        with self._lock:
            if self._state == _CANCELLED:
                raise CancelledError
            if self._state != _FINISHED:
                raise InvalidStateError("Exception is not set.")
            self._log_traceback = False
            return self._exception

    def __await__(self) -> Generator[Any, None, T_co]:
        if not self.done():
            self._asyncio_future_blocking = True
            yield self  # This tells Task to wait for completion.
        assert self.done(), "yield from wasn't used with future"
        return self.result()  # May raise too.


class DeferredStartTask:
    """
    A :class:`Task`-like object that delays starting its coroutine until the :meth:`start` method is
    called.
    """

    _asyncio_future_blocking = False

    def __init__(self, cb: Callable[[], Future], start=False, name=None):
        if not callable(cb):
            raise ValueError("cb must be callable")

        if start:
            self._future = cb()
            self._callback = None
        else:
            self._future = get_event_loop().create_future()
            self._callback = cb

        # This is required to be compatible with the interface of a Future.
        self._loop = self._future._loop
        self._name = name or repr(cb)

    def start(self) -> None:
        """
        Actually invoke the underlying function (but only if it hasn't been invoked already).
        """
        cb = self._callback
        if cb is not None:
            self._callback = None
            propagate(from_=ensure_future(cb()), to=self._future)

    def started(self) -> bool:
        return self._callback is None

    def cancelled(self) -> bool:
        return self._future.cancelled()

    def cancel(self) -> bool:
        self._callback = None
        return self._future.cancel()

    def result(self) -> Any:
        return self._future.result()

    def done(self) -> bool:
        return self._future.done()

    def exception(self) -> Optional[BaseException]:
        return self._future.exception()

    def add_done_callback(self, callback: "Callable[[Future], Any]", context=None) -> None:
        if context is None:
            self._future.add_done_callback(callback)
        else:
            self._future.add_done_callback(callback, context=context)  # type: ignore

    def __await__(self):
        return self._future.__await__()

    def __repr__(self):
        state = self._future._state if self.started() else "NOT_STARTED"
        return f"DeferredStartTask({self._name!r}, state={state})"


@no_type_check
def get_running_loop() -> Optional[AbstractEventLoop]:
    warnings.warn(
        "get_running_loop() is deprecated, as support for Python 3.6 will be dropped in dazl 8. "
        "Use Python 3.7+'s asyncio.get_running_loop() instead.",
        DeprecationWarning,
    )
    try:
        from asyncio import get_running_loop

        try:
            return get_running_loop()
        except RuntimeError:
            return None
    except ImportError:
        # noinspection PyProtectedMember
        from asyncio import _get_running_loop

        return _get_running_loop()


# noinspection PyDeprecation
def safe_create_future():
    warnings.warn(
        "safe_create_future() is deprecated; there is no planned replacement.",
        DeprecationWarning,
        stacklevel=2,
    )
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        loop = get_running_loop()
    return loop.create_future() if loop is not None else ContextFreeFuture()


@no_type_check
def named_gather(name: str, *awaitables, return_exceptions=False):
    g = gather(*awaitables, return_exceptions=return_exceptions)
    g.__repr__ = staticmethod(lambda _: f"<Gather {name}>")
    return g


class Signal:
    """
    A simple guard that "wakes" up a coroutine from another coroutine.
    """

    # noinspection PyDeprecation
    def __init__(self):
        warnings.warn(
            "Signal is deprecated; there is no planned replacement",
            DeprecationWarning,
            stacklevel=2,
        )
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", DeprecationWarning)
            self._fut = safe_create_future()

    def notify_all(self) -> None:
        """
        Schedule a notification to all coroutines that are `wait`ing.
        """
        if self._fut is not None:
            self._fut.set_result(None)
            self._fut = None

    # noinspection PyDeprecation
    def wait(self) -> "Awaitable[None]":
        if self._fut is None:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", DeprecationWarning)
                self._fut = safe_create_future()
        return self._fut


class UnsettableEventLoopPolicy(AbstractEventLoopPolicy):
    def get_event_loop(self) -> AbstractEventLoop:
        raise Exception("Called get_event_loop")

    def set_event_loop(self, loop: Optional[AbstractEventLoop]) -> None:
        raise Exception("Called set_event_loop")

    def new_event_loop(self) -> AbstractEventLoop:
        raise Exception("Called new_event_loop")

    def get_child_watcher(self) -> Any:
        raise Exception("get_child_watcher")

    def set_child_watcher(self, watcher: Any) -> None:
        raise Exception("set_child_watcher")



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import os
from threading import current_thread, main_thread
from typing import Optional

from ..util.enum import OrderedEnum


class RunLevel(OrderedEnum):
    """
    The possible running modes for an :class:`Invoker`. Invokers generally progress from
    lower-numbered run levels to higher-numbered run levels.
    """

    RUN_FOREVER = 0
    RUN_UNTIL_IDLE = 1
    TERMINATE_GRACEFULLY = 2
    TERMINATE_IMMEDIATELY = 3
    STOPPED = 4


def validate_install_signal_handlers(user_install_signal_handlers: "Optional[bool]" = None) -> bool:
    if user_install_signal_handlers is None:
        # The default behavior for signal handling is to suppress if we are already running in a
        # background thread or we are being run within unit tests.
        return "PYTEST_CURRENT_TEST" not in os.environ and current_thread() is main_thread()

    if user_install_signal_handlers and current_thread() is not main_thread():
        # The user WANTS signal handlers, but we know the registration call will fail
        raise RuntimeError("cannot install signal handlers when not on the main thread")

    return user_install_signal_handlers



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
This module contains the types needed to submit commands to and read events from a
Daml `gRPC Ledger API <https://docs.daml.com/app-dev/ledger-api.html>`_ or
`HTTP JSON API <https://docs.daml.com/json-api/index.html>`_.
"""
import sys

from .api_types import (
    ArchiveEvent,
    Boundary,
    Command,
    CreateAndExerciseCommand,
    CreateCommand,
    CreateEvent,
    Event,
    EventOrBoundary,
    ExerciseByKeyCommand,
    ExerciseCommand,
    ExerciseResponse,
    PartyInfo,
    SubmitResponse,
)

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable


__all__ = [
    "aio",
    "ArchiveEvent",
    "Boundary",
    "Command",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "CreateEvent",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "ExerciseResponse",
    "PartyInfo",
    "PackageService",
    "Connection",
    "QueryStream",
]


# noinspection PyShadowingNames
def connect(*, blocking=False, **kwargs):
    """
    Create a connection from the supplied parameters.

    See the `documentation for this function
    <https://digital-asset.github.io/dazl-client/dazl.ledger.html#dazl.ledger.connect>`_ for more
    details on the parameters it takes and how values are defaulted.
    """
    from .config import Config
    from .grpc.conn_aio import Connection as GrpcConnection

    if blocking:
        raise ValueError("blocking connections are not currently supported")

    cfg = Config.create(**kwargs)
    return GrpcConnection(cfg)


class PackageService(Protocol):
    """
    Protocol that describe a service that provides package information. The :class:`Connection`
    protocol extends this interface.
    """

    def get_package(self, __package_id):
        """
        Given a package ID, fetch the binary data for the corresponding DALF.

        :param __package_id:
            The package ID of the DALF to retrieve.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :return:
            The byte array contents of the DALF associated with the package ID.
        """
        raise NotImplementedError

    def list_package_ids(self):
        """
        Fetch a list of all known package IDs.
        """
        raise NotImplementedError


@runtime_checkable
class Connection(PackageService, Protocol):
    """
    Protocol that describes a connection to a ledger. You will typically work with the more specific
    protocols :class:`dazl.ledger.aio.Connection` or :class:`dazl.ledger.blocking.Connection` that
    are tailored towards connections with ``asyncio`` or thread-blocking semantics, respectively.
    """

    @property
    def config(self):
        """
        Get the configuration that created this connection.

        Under certain circumstances (notably, OAuth token refreshing), fields of this configuration
        object are modifiable after the connection has been created; see the documentation for
        :class:`dazl.ledger.config.Config` for more information.
        """
        raise NotImplementedError

    @property
    def codec(self):
        """
        Get the object that provides conversion to over-the-wire types for this connection.

        The specific object here is transport-specific.
        """
        raise NotImplementedError

    def open(self):
        """
        Prepares the connection for being used, including possibly fetching the ledger ID if it is
        not yet known.

        On *asynchronous* connections, this method is a coroutine.

        On *blocking* connections, this method blocks the current thread until a connection is
        ready.
        """
        raise NotImplementedError

    def close(self):
        """
        Close the connection.

        On _asynchronous_ connections, this method is a coroutine.

        On _blocking_ connections, this method blocks the current thread until the connection is
        closed and all network activity has stopped.
        """
        raise NotImplementedError

    def submit(self, __commands, *, workflow_id=None, command_id=None):
        """
        Submit one or more commands to the Ledger API.

        You should generally prefer trying to use :meth:`create`, :meth:`exercise`,
        :meth:`exercise_by_key`, or :meth:`create_and_exercise`, as they are available over both
        the gRPC Ledger API and HTTP JSON API; additionally those methods can provide more
        information about what happened.

        This method can be used to submit multiple disparate commands as a single transaction, but
        if you find yourself needing to do this, you may want to consider moving more of your logic
        into Daml so that only a single command is needed from the outside in order to satisfy your
        use case.

        On *asynchronous* connections, this method is a coroutine.

        On *blocking* connections, this method blocks the current thread until the connection is
        closed and all network activity has stopped.

        :param __commands:
            A command or sequence of commands to submit to the ledger.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        """
        raise NotImplementedError

    def create(self, __template_id, __payload, *, workflow_id, command_id):
        """
        Create a contract for a given template.

        :param __template_id:
            The template of the contract to be created.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __payload:
            Template arguments for the contract to be created.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :return:
            The :class:`CreateEvent` that represents the contract that was successfully created.
        """
        raise NotImplementedError

    def exercise(self, __contract_id, __choice_name, __argument, *, workflow_id, command_id):
        """
        Exercise a choice on a contract identified by its contract ID.

        :param __contract_id:
            The contract ID of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __choice_name:
            The name of the choice to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __argument:
            The choice arguments. Can be omitted for choices that take no argument.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        raise NotImplementedError

    def create_and_exercise(
        self,
        __template_id,
        __payload,
        __choice_name,
        __argument=None,
        *,
        workflow_id=None,
        command_id=None
    ):
        """
        Exercise a choice on a newly-created contract, in a single transaction.

        :param __template_id:
            The template of the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __payload:
            Template arguments for the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __choice_name:
            The name of the choice to exercise (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __argument:
            The choice arguments. Can be omitted for choices that take no argument (positional
            argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        raise NotImplementedError

    def exercise_by_key(
        self, __template_id, __choice_name, __key, __argument, *, workflow_id=None, command_id=None
    ):
        """
        Exercise a choice on a contract identified by its contract key.

        :param __template_id:
            The template of the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __choice_name:
            The name of the choice to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __key:
            The key of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __argument:
            The choice arguments. Can be omitted for choices that take no argument.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        raise NotImplementedError

    def archive(self, __contract_id, *, workflow_id=None, command_id=None):
        """
        Archive a choice on a contract identified by its contract ID.

        :param __contract_id:
            The contract ID of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        raise NotImplementedError

    def archive_by_key(self, __template_id, __key, *, workflow_id=None, command_id=None):
        """
        Exercise a choice on a contract identified by its contract key.

        :param __template_id:
            The template of the contract to be created (positional argument only).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __key:
            The key of the contract to exercise.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param workflow_id:
            An optional workflow ID.
        :param command_id:
            An optional command ID. If unspecified, a random one will be created.
        :return:
            The return value of the choice, together with a list of events that occurred as a result
            of exercising the choice.
        """
        raise NotImplementedError

    def query(self, __template_id="*", __query=None):
        """
        Return the create events from the active contract set service as a stream.

        If you find yourself repeatedly calling :meth:`query` or :meth:`query_many` over the same
        set of templates, you may want to consider :class:`ACS` instead, which is a utility class
        that helps you maintain a "live" state of the ACS.

        :param __template_id:
            The name of the template for which to fetch contracts.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __query:
            A filter to apply to the set of returned contracts.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :return:
            A query stream that iterates over the :class:`CreateEvent` objects of the Active
            Contract Set, and terminates.
        """
        raise NotImplementedError

    def query_many(self, *q):
        """
        Return the create events from the active contract set service as a stream.

        If you find yourself repeatedly calling :meth:`query` or :meth:`query_many` over the same
        set of templates, you may want to consider :class:`ACS` instead, which is a utility class
        that helps you maintain a "live" state of the ACS.

        Examples:

            # Return a stream that returns a snapshot of the complete ACS
            stream = conn.query_many()

            # Return a stream that iterates over the Magic template in the So.Much module
            stream = conn.query_many("So.Much:Magic")

            # Return a stream that iterates over two templates
            # (can either be called with multiple parameters or as a single list)
            stream = conn.query_many("So.Much:Magic", "So.Much:Awesome")
            stream = conn.query_many(["So.Much:Magic", "So.Much:Awesome"])

            # Return a stream that iterates over Magic templates
            # with a field `power` having a value of 100
            stream = conn.query_many({"So.Much:Magic": {"power": 100}})
        """
        raise NotImplementedError

    def stream(self, __template_id="*", __query=None, *, offset=None):
        """
        Stream create/archive events.

        When ``offset`` is ``None``, create events from the active contract set are returned first,
        followed by a continuous stream of updates (creates/archives).

        Otherwise, ``offset`` can be supplied to resume a stream from a prior point where a
        ``Boundary`` was returned from a previous stream.

        :param __template_id:
            The name of the template for which to fetch contracts.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param __query:
            A filter to apply to the set of returned contracts. Note that this does not filter
            :class:`ArchiveEvent`; readers of the stream MUST be able to cope with "mismatched"
            archives that come from the result of applying a filter.

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        :param offset:
            An optional offset at which to start receiving events. If ``None``, start from the
            beginning.
        """
        raise NotImplementedError

    def stream_many(self, *q, offset=None):
        """
        Stream create/archive events from more than one template ID in the same stream.

        When ``offset`` is ``None``, create events from the active contract set are returned first,
        followed by a continuous stream of updates (creates/archives).

        Otherwise, ``offset`` can be supplied to resume a stream from a prior point where a
        ``Boundary`` was returned from a previous stream.

        Examples:

            # Return a stream that returns a snapshot of the complete ACS
            stream = conn.stream_many()

            # Return a stream that iterates over the Magic template in the So.Much module
            stream = conn.stream_many("So.Much:Magic")

            # Return a stream that iterates over two templates
            # (can either be called with multiple parameters or as a single list)
            stream = conn.stream_many("So.Much:Magic", "So.Much:Awesome")
            stream = conn.stream_many(["So.Much:Magic", "So.Much:Awesome"])

            # Return a stream that iterates over Magic templates
            # with a field `power` having a value of 100
            stream = conn.stream_many({"So.Much:Magic": {"power": 100}})
        """
        raise NotImplementedError

    def allocate_party(self, *, identifier_hint=None, display_name=None):
        """
        Allocate a new party.

        This method requires that the connection be created with ``admin=true``, or a token that
        contains an ``admin=true`` claim.

        :param identifier_hint:
            A hint to the backing participant. Note that the Daml ledger is free to ignore this
            hint entirely.
        :type identifier_hint: str
        :param display_name:
            A human-readable name that corresponds to the allocated ``Party``.
        :type display_name: str
        :return:
            Information about the allocated ``Party``.
        """
        raise NotImplementedError

    def list_known_parties(self):
        """
        Return a list of :class:`PartyInfo` for all parties on the ledger.
        """
        raise NotImplementedError

    def upload_package(self, __contents):
        """
        Upload a byte array as a DAR file to the remote Daml ledger.

        This method requires that the connection be created with ``admin=true``, or a token that
        contains an ``admin=true`` claim.

        :param __contents:
            The contents of a DAR file (for example, as obtained from running ``daml build``).

            Note that future versions of dazl reserve the right to rename this parameter name at any
            time; it should be passed in as a positional parameter and never by name.
        """
        raise NotImplementedError


class QueryStream(Protocol):
    """
    Protocol for classes that provide for reading from a stream of events from a Daml ledger. Like
    :class:`Connection`, there are async query streams (:class:`dazl.ledger.aio.QueryStream`) and
    blocking query streams (:class:`dazl.ledger.blocking.QueryStream`).
    """

    def on_create(self, *args):
        """
        Register a callback that is triggered whenever a :class:`CreateEvent` is read through the
        stream.
        """
        raise NotImplementedError

    def on_archive(self, *args):
        """
        Register a callback that is triggered whenever a :class:`ArchiveEvent` is read through the
        stream.
        """
        raise NotImplementedError

    def on_boundary(self, *args):
        """
        Register a callback that is triggered whenever a :class:`Boundary` is read through the
        stream.
        """
        raise NotImplementedError

    def close(self):
        """
        Close and dispose of any resources used by this stream.

        This can be called at any time to end a stream prematurely, and is the only way to abort
        a stream returned by :meth:`Connection.stream`/:meth:`Connection.stream_many`.
        """
        raise NotImplementedError

    def run(self):
        """
        Consume the stream without returning any values.

        This method is useful if you use :meth:`on_create`, :meth:`on_archive`, or
        :meth:`on_boundary` to register callbacks for events.

        On _asynchronous_ connections, this method is a coroutine that terminates once the stream
        has ended. For streams returned by :meth:`Connection.stream`/:meth:`Connection.stream_many`,
        the stream is only ended by :meth:`close`.

        On _blocking_ connections, this method blocks until the stream has ended.

        For streams returned by :meth:`Connection.stream`/:meth:`Connection.stream_many`, the
        method blocks indefinitely, and is only closed if :meth:`close` is called or the
        server disconnects.
        """
        raise NotImplementedError

    def creates(self):
        """
        Return a stream of :class:`CreateEvent` objects. This will include the contracts of the
        Active Contract Set, as well as create events over subsequent transactions.

        On _asynchronous_ connections, this method is an asynchronous iterator: use ``async for``
        to iterate over its contents.

        On _blocking_ connections, this method a blocking iterator: use ``for`` to iterate over its
        contents.

        For streams returned by :meth:`Connection.stream`/:meth:`Connection.stream_many`, the
        iterator remains open indefinitely, and is only closed if :meth:`close` is called or the
        server disconnects.
        """
        raise NotImplementedError

    def events(self):
        """
        Return a stream of :class:`CreateEvent` and :class:`ArchiveEvent` objects. This will include
        the initial contracts of the Active Contract Set, as well as create and archive events over
        subsequent transactions. You may see :class:`ArchiveEvent` objects that had no
        :class:`CreateEvent`s predecessor; this will typically happen with event filtering (which
        can only effectively filter over :class:`CreateEvent`

        On _asynchronous_ connections, this method is an asynchronous iterator: use ``async for``
        to iterate over its contents.

        On _blocking_ connections, this method a blocking iterator: use ``for`` to iterate over its
        contents.

        For streams returned by :meth:`Connection.stream`/:meth:`Connection.stream_many`, the
        iterator remains open indefinitely, and is only closed if :meth:`close` is called or the
        server disconnects.
        """
        raise NotImplementedError

    def items(self):
        """
        Return a stream of :class:`CreateEvent`, :class:`ArchiveEvent`, and :class:`Boundary`
        objects. This will include the initial contracts of the Active Contract Set, as well as
        create and archive events over subsequent transactions, and :class:`Boundary` objects that
        describe offsets where an interrupted stream can be resumed without having to re-read the
        initial contracts of the Active Contract Set.

        On _asynchronous_ connections, this method is an asynchronous iterator: use ``async for``
        to iterate over its contents.

        On _blocking_ connections, this method a blocking iterator: use ``for`` to iterate over its
        contents.

        For streams returned by :meth:`Connection.stream`/:meth:`Connection.stream_many`, the
        iterator remains open indefinitely, and is only closed if :meth:`close` is called or the
        server disconnects.
        """
        raise NotImplementedError



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Collection,
    Mapping,
    NoReturn,
    Optional,
    Sequence,
    Union,
)

from ..damlast.daml_lf_1 import TypeConName
from ..damlast.lookup import parse_type_con_name
from ..prim import ContractData, ContractId, Party
from ..util.typing import safe_cast

__all__ = [
    "ArchiveEvent",
    "Boundary",
    "Command",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "CreateEvent",
    "Event",
    "EventOrBoundary",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "ExerciseResponse",
    "PartyInfo",
    "SubmitResponse",
]


class Command:
    """
    Base class for write-side commands.

    This class provides no functionality on its own.
    """

    def __setattr__(self, key, value) -> NoReturn:
        """
        Raise :class:`AttributeError`; instances of this class are immutable.
        """
        raise AttributeError("Command instances are read-only")


class CreateCommand(Command):
    """
    A command that creates a contract without any predecessors.
    """

    __slots__ = ("_template_id", "_payload")
    if TYPE_CHECKING:
        _template_id: TypeConName
        _payload: ContractData

    def __init__(self, template_id: Union[str, TypeConName], payload: ContractData):
        """
        Initialize a :class:`CreateCommand`.

        :param template_id:
            The template of the contract to be created.
        :param payload:
            The template arguments for the contract to be created.
        """
        object.__setattr__(self, "_template_id", validate_template_id(template_id))
        object.__setattr__(self, "_payload", payload)

    @property
    def template_id(self) -> TypeConName:
        """
        Return the template of the contract to be created.
        """
        return self._template_id

    @property
    def payload(self) -> Mapping[str, Any]:
        """
        Return the template arguments for the contract to be created.
        """
        return self._payload

    def __eq__(self, other: "Any") -> bool:
        return (
            isinstance(other, CreateCommand)
            and self.template_id == other.template_id
            and self.payload == other.payload
        )

    def __repr__(self) -> str:
        return f"CreateCommand({self.template_id}, {self.payload})"


class CreateAndExerciseCommand(Command):
    """
    A command that exercises a choice on a newly-created contract in a single transaction.

    Instead of creating an instance of this command and submitting it with :meth:`Connection.submit`,
    consider using :meth:`Connection.create_and_exercise` instead, which also gives you access to
    the result of exercising the choice.
    """

    __slots__ = ("_template_id", "_payload", "_choice", "_argument")
    if TYPE_CHECKING:
        _template_id: TypeConName
        _payload: ContractData
        _choice: str
        _argument: Any

    def __init__(
        self,
        template_id: Union[str, TypeConName],
        payload: ContractData,
        choice: str,
        argument: Optional[Any] = None,
    ):
        """
        Initialize a :class:`CreateAndExerciseCommand`.

        :param template_id:
            The template of the contract to be created.
        :param payload:
            The template arguments for the contract to be created.
        :param choice:
            The choice to exercise.
        :param argument:
            The choice arguments. Can be omitted for choices that take no arguments.
        """
        object.__setattr__(self, "_template_id", validate_template_id(template_id))
        object.__setattr__(self, "_payload", payload)
        object.__setattr__(self, "_choice", choice)
        object.__setattr__(self, "_argument", dict(argument) if argument is not None else dict())

    @property
    def template_id(self) -> TypeConName:
        """
        The template of the contract to be created.
        """
        return self._template_id

    @property
    def payload(self) -> ContractData:
        """
        The template arguments for the contract to be created.
        """
        return self._payload

    @property
    def choice(self) -> str:
        """
        The choice to exercise.
        """
        return self._choice

    @property
    def argument(self) -> Any:
        """
        The choice arguments.
        """
        return self._argument

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, CreateAndExerciseCommand)
            and self.template_id == other.template_id
            and self.payload == other.payload
            and self.choice == other.choice
            and self.argument == other.argument
        )

    def __repr__(self) -> str:
        return f"CreateAndExerciseCommand({self.template_id}, {self.payload}, {self.choice!r}, {self.argument})"


class ExerciseCommand(Command):
    """
    A command that exercises a choice on a contract identified by its contract ID.

    Instead of creating an instance of this command and submitting it with :meth:`Connection.submit`,
    consider using :meth:`Connection.exercise` instead, which also gives you access to the result of
    exercising the choice.
    """

    __slots__ = ("_choice", "_contract_id", "_argument")
    if TYPE_CHECKING:
        _choice: str
        _contract_id: ContractId
        _argument: Optional[Any]

    def __init__(self, contract_id: ContractId, choice: str, argument: Optional[Any] = None):
        """
        Initialize an :class:`ExerciseCommand`.

        :param contract_id:
            The contract ID of the contract to exercise.
        :param choice:
            The choice to exercise.
        :param argument:
            The choice arguments. Can be omitted for choices that take no arguments.
        """
        object.__setattr__(self, "_choice", safe_cast(str, choice))
        object.__setattr__(self, "_contract_id", safe_cast(ContractId, contract_id))
        object.__setattr__(self, "_argument", dict(argument) if argument is not None else dict())

    @property
    def contract_id(self) -> ContractId:
        """
        The contract ID of the contract to exercise.
        """
        return self._contract_id

    @property
    def choice(self) -> str:
        """
        The choice to exercise.
        """
        return self._choice

    @property
    def argument(self) -> Any:
        """
        The choice arguments.
        """
        return self._argument

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ExerciseCommand)
            and self.choice == other.choice
            and self.contract_id == other.contract_id
            and self.argument == other.argument
        )

    def __repr__(self):
        return f"ExerciseCommand({self.choice!r}, {self.contract_id}, {self.argument})"


class ExerciseByKeyCommand(Command):
    """
    A command that exercises a choice on a contract identified by its contract key.

    Instead of creating an instance of this command and submitting it with :meth:`Connection.submit`,
    consider using :meth:`Connection.exercise_by_key` instead, which also gives you access to the
    result of exercising the choice.
    """

    __slots__ = ("_template_id", "_key", "_choice", "_argument")
    if TYPE_CHECKING:
        _template_id: TypeConName
        _key: Any
        _choice: str
        _argument: Optional[Any]

    def __init__(
        self,
        template_id: Union[str, TypeConName],
        key: Any,
        choice: str,
        argument: Optional[Any] = None,
    ):
        """
        Initialize an :class:`ExerciseByKeyCommand`.

        :param template_id:
            The contract template type.
        :param key:
            The contract key of the contract to exercise.
        :param choice:
            The choice to exercise.
        :param argument:
            The choice arguments. Can be omitted for choices that take no arguments.
        """
        object.__setattr__(self, "_template_id", validate_template_id(template_id))
        object.__setattr__(self, "_key", key)
        object.__setattr__(self, "_choice", choice)
        object.__setattr__(self, "_argument", dict(argument) if argument is not None else dict())

    @property
    def template_id(self) -> TypeConName:
        """
        The contract template type.
        """
        return self._template_id

    @property
    def key(self) -> Any:
        """
        The contract key of the contract to exercise.
        """
        return self._key

    @property
    def choice(self) -> str:
        """
        The choice to exercise.
        """
        return self._choice

    @property
    def argument(self) -> Any:
        """
        The choice arguments.
        """
        return self._argument

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, ExerciseByKeyCommand)
            and self.template_id == other.template_id
            and self.key == other.key
            and self.choice == other.choice
            and self.argument == other.argument
        )

    def __repr__(self):
        return f"ExerciseByKeyCommand({self.template_id}, {self.key}, {self.choice!r}, {self.argument})"


class CreateEvent:
    """
    An event that indicates a newly-created contract.
    """

    __slots__ = (
        "_contract_id",
        "_payload",
        "_signatories",
        "_observers",
        "_agreement_text",
        "_key",
    )
    if TYPE_CHECKING:
        _contract_id: ContractId
        _payload: ContractData
        _signatories: AbstractSet[Party]
        _observers: AbstractSet[Party]
        _agreement_text: Optional[str]
        _key: Optional[Any]

    def __init__(
        self,
        contract_id: ContractId,
        payload: ContractData,
        signatories: Collection[Party],
        observers: Collection[Party],
        agreement_text: Optional[str],
        key: Optional[Any],
    ):
        object.__setattr__(self, "_contract_id", contract_id)
        object.__setattr__(self, "_payload", payload)
        object.__setattr__(self, "_signatories", frozenset(signatories))
        object.__setattr__(self, "_observers", frozenset(observers))
        object.__setattr__(self, "_agreement_text", agreement_text)
        object.__setattr__(self, "_key", key)

    @property
    def contract_id(self) -> ContractId:
        """
        The ID of the created contract.
        """
        return self._contract_id

    @property
    def payload(self) -> ContractData:
        """
        The `parameters <https://docs.daml.com/daml/reference/templates.html#template-parameters>`_
        that were used to create the contract.
        """
        return self._payload

    @property
    def signatories(self) -> AbstractSet[Party]:
        """
        The `signatories <https://docs.daml.com/daml/reference/templates.html#signatory-parties>`_
        for this contract as specified by the template.
        """
        return self._signatories

    @property
    def observers(self) -> AbstractSet[Party]:
        """
        The `observers <https://docs.daml.com/daml/reference/templates.html#observers>`_ for this
        contract as specified explicitly by the template or implicitly as choice controllers.
        """
        return self._observers

    @property
    def agreement_text(self) -> Optional[str]:
        """
        The `agreement text <https://docs.daml.com/daml/reference/templates.html#agreements>`_ of
        the contract.
        """
        return self._agreement_text

    @property
    def key(self) -> Optional[Any]:
        """
        The `key <https://docs.daml.com/daml/reference/templates.html#contract-keys-and-maintainers>`_
        of the contract, if defined.
        """
        return self._key

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, CreateEvent)
            and self.contract_id == other.contract_id
            and self.payload == other.payload
            and self.signatories == other.signatories
            and self.observers == other.observers
            and self.agreement_text == other.agreement_text
            and self.key == other.key
        )


class ArchiveEvent:
    """
    An event that indicates a contract was archived.
    """

    __slots__ = ("_contract_id",)
    if TYPE_CHECKING:
        _contract_id: ContractId

    def __init__(self, contract_id: ContractId):
        object.__setattr__(self, "_contract_id", contract_id)

    @property
    def contract_id(self) -> ContractId:
        """
        The contract ID of the archived contract.
        """
        return self._contract_id


Event = Union[CreateEvent, ArchiveEvent]


class Boundary:
    """
    An event that indicates a boundary in a query stream where events can be resumed.
    """

    __slots__ = ("_offset",)
    if TYPE_CHECKING:
        _offset: Optional[str]

    def __init__(self, offset: Optional[str]):
        object.__setattr__(self, "_offset", offset)

    @property
    def offset(self) -> Optional[str]:
        """
        The offset at which this boundary occurred.

        If this is ``None``, that indicates that an active contract set was requested, but the
        ledger is completely empty.
        """
        return self._offset

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, Boundary) and self.offset == other.offset

    def __hash__(self):
        return hash(self._offset)

    def __repr__(self):
        return f"Boundary({self._offset!r})"


EventOrBoundary = Union[Event, Boundary]


class ExerciseResponse:
    """
    Returned when directly exercising a choice using :meth:`Connection.create_and_exercise`,
    :meth:`Connection.exercise`, or :meth:`Connection.exercise_by_key`.
    """

    __slots__ = ("_result", "_events")
    if TYPE_CHECKING:
        _result: Optional[Any]
        _events: Sequence[Union[CreateEvent, ArchiveEvent]]

    def __init__(self, result: Optional[Any], events: Sequence[Union[CreateEvent, ArchiveEvent]]):
        object.__setattr__(self, "_result", result)
        object.__setattr__(self, "_events", tuple(events))

    @property
    def result(self) -> Optional[Any]:
        """
        The return value of the choice.
        """
        return self._result

    @property
    def events(self) -> Sequence[Union[CreateEvent, ArchiveEvent]]:
        """
        All of the events that occurred as a result of exercising the choice, including the archive
        event for the contract if the choice is consuming (or otherwise archives it as part of its
        execution).
        """
        return self._events

    def __repr__(self):
        return f"ExerciseResponse(result={self.result}, events={self.events})"


SubmitResponse = Union[None, CreateEvent, ExerciseResponse]


class PartyInfo:
    """
    Full information about a ``Party``.
    """

    __slots__ = ("_party", "_display_name", "_is_local")
    if TYPE_CHECKING:
        _party: Party
        _display_name: str
        _is_local: bool

    def __init__(self, party: Party, display_name: str, is_local: bool):
        object.__setattr__(self, "_party", party)
        object.__setattr__(self, "_display_name", display_name)
        object.__setattr__(self, "_is_local", is_local)

    @property
    def party(self) -> "Party":
        """
        The stable unique identifier of a Daml ``Party``.
        """
        return self._party

    @property
    def display_name(self) -> str:
        """
        The human-readable name associated with the ``Party``.
        """
        return self._display_name

    @property
    def is_local(self) -> bool:
        """
        Indicates if the ``Party`` is hosted by the backing participant.
        """
        return self._is_local


def validate_template_id(value: Union[str, TypeConName]) -> TypeConName:
    if isinstance(value, TypeConName):
        return value
    else:
        return parse_type_con_name(value)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from datetime import timedelta
import ipaddress
from logging import Logger, getLogger
import os
from reprlib import repr
import sys
from types import MappingProxyType
from typing import TYPE_CHECKING, Optional
from urllib.parse import urlparse
import warnings

from ...prim import TimeDeltaLike, to_timedelta
from .exc import ConfigError, ConfigWarning

if sys.version_info >= (3, 8):
    from typing import Protocol, runtime_checkable
else:
    from typing_extensions import Protocol, runtime_checkable


__all__ = ["URLConfig", "create_url", "KNOWN_SCHEME_PORTS", "DEFAULT_CONNECT_TIMEOUT"]

if TYPE_CHECKING:
    # We refer to the Config class in a docstring and
    # without this import, Sphinx can't resolve the reference
    # noinspection PyUnresolvedReferences
    from . import Config

DEFAULT_CONNECT_TIMEOUT = timedelta(seconds=30)

# The set of schemes we understand, and the known ports that map to those schemes. The first port
# in the list is the default value for the scheme.
KNOWN_SCHEME_PORTS = MappingProxyType(
    {"http": (80, 7575, 8080), "https": (443, 8443), "grpc": (6865,), "grpcs": ()}
)

# some environment variables that we frequently refer to
DAML_LEDGER_URL = "DAML_LEDGER_URL"
DAML_LEDGER_HOST = "DAML_LEDGER_HOST"
DAML_LEDGER_PORT = "DAML_LEDGER_PORT"
DAML_LEDGER_SCHEME = "DAML_LEDGER_SCHEME"


def create_url(
    *,
    url: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    scheme: Optional[str] = None,
    connect_timeout: Optional[TimeDeltaLike] = None,
    use_http_proxy: Optional[bool] = None,
    logger: Optional[Logger] = None,
):
    """
    Create an instance of :class:`URLConfig`, possibly with values taken from environment variables,
    or defaulted if otherwise unspecified.

    See :meth:`Config.create` for a more detailed description of these parameters.
    """
    if logger is None:
        logger = getLogger("dazl.conn")

    if not url and not host and not port and not scheme:
        # use environment variables to provide default values
        url = os.getenv(DAML_LEDGER_URL)
        host = os.getenv(DAML_LEDGER_HOST)
        port_str = os.getenv(DAML_LEDGER_PORT)
        port = int(port_str) if port_str else None
        scheme = os.getenv(DAML_LEDGER_SCHEME)
        if host or port or scheme:
            if url:
                logger.error("Found conflicting environment variables:")
                logger.error("     %s=%s", DAML_LEDGER_URL, url)
                logger.error("     %s=%s", DAML_LEDGER_HOST, host)
                logger.error("     %s=%s", DAML_LEDGER_PORT, port)
                logger.error("     %s=%s", DAML_LEDGER_SCHEME, scheme)
                logger.error(
                    f"Specify ONLY either {DAML_LEDGER_URL} OR "
                    "{DAML_LEDGER_HOST}/{DAML_LEDGER_PORT}/{DAML_LEDGER_SCHEME}"
                )
            else:
                logger.info("Using URL configuration from the environment:")
                logger.info("     %s=%s", DAML_LEDGER_HOST, host)
                logger.info("     %s=%s", DAML_LEDGER_PORT, port)
                logger.info("     %s=%s", DAML_LEDGER_SCHEME, scheme)
        elif url:
            logger.info("Using URL configuration from the environment:")
            logger.info("     %s=%s", DAML_LEDGER_URL, url)

    if not url and not host and not port and not scheme:
        # if no values are supplied _and_ no environment variables are specified either, then
        # fall back to default values
        logger.info(
            'Configuring a connection to "localhost" because neither url/host/port/scheme nor '
            f"the environment variables {DAML_LEDGER_URL}, {DAML_LEDGER_HOST}, {DAML_LEDGER_PORT}, "
            f"or {DAML_LEDGER_SCHEME} are defined"
        )
        url = "localhost"

    if url:
        if host or port or scheme:
            raise ValueError("url or host/port/scheme must be specified, but not both")
        url = sanitize_url(url)
    else:
        url = build_url(host, port, scheme)

    if use_http_proxy is None:
        hostname = urlparse(url).hostname
        if hostname is None:
            # this shouldn't have happened because we validate hostnames above
            raise ValueError("hostname of None unexpected here")

        use_http_proxy = not is_localhost(hostname)

    logger.debug("Building a URL configuration:")
    logger.debug("    url=%s", url)
    logger.debug("    connect_timeout=%s", connect_timeout)
    logger.debug("    use_http_proxy=%s", use_http_proxy)

    return SimpleURLConfig(
        url=url,
        connect_timeout=to_timedelta(connect_timeout)
        if connect_timeout is not None
        else DEFAULT_CONNECT_TIMEOUT,
        use_http_proxy=use_http_proxy,
    )


@runtime_checkable
class URLConfig(Protocol):
    """
    Configuration parameters for the remote host, including basic connection parameters that do not
    belong in :class:`AccessConfig`.
    """

    @property
    def url(self) -> str:
        """
        The full URL to connect to, including a protocol (scheme), host, and port.
        """
        raise NotImplementedError

    @property
    def use_http_proxy(self) -> bool:
        """
        Whether to allow the use of HTTP proxies.
        """
        raise NotImplementedError

    @property
    def connect_timeout(self) -> timedelta:
        """
        How long to wait for a connection before giving up.

        The default is 30 seconds.
        """
        raise NotImplementedError


class SimpleURLConfig:
    """
    Trivial implementation of the :class:`URLConfig` protocol.
    """

    def __init__(self, url: str, connect_timeout: timedelta, use_http_proxy: bool):
        self.url = url
        self.connect_timeout = connect_timeout
        self.use_http_proxy = use_http_proxy


def sanitize_url(url: str) -> str:
    """
    Perform some basic sanitization on a URL string (see :meth:`build_url`):
     * Supply a default protocol according to our general rules
     * Supply a port, even if one wasn't specified
     * Strip out any trailing slashes

    >>> sanitize_url("somewhere:1000")
    '//somewhere:1000'

    >>> sanitize_url("http://somewhere:1000")
    'http://somewhere:1000'

    >>> sanitize_url("http://somewhere:1000/")
    'http://somewhere:1000/'
    """
    first_slash = url.find("/")
    if first_slash == -1 or first_slash != url.find("//"):
        url = "//" + url

    original = urlparse(url)
    sanitized = urlparse(build_url(original.hostname, original.port, original.scheme))
    return original._replace(
        netloc=f"{sanitized.hostname}:{sanitized.port}", scheme=sanitized.scheme
    ).geturl()


def build_url(host: Optional[str], port: Optional[int], scheme: Optional[str]) -> str:
    """
    Build a URL from host/port/scheme components.

    :param host:
        The host to connect to. If unspecified, ``"localhost"`` is assumed.
    :param port:
        The port to connect to. If provided, must be a valid port number
        (between 1 and 65535 inclusive).
    :param scheme:
        The scheme to use (must be ``"http"``, ``"https"``, ``"grpc"``, ``"grpcs"`` or ``None``).
    :raises ConfigError: One or more of the passed parameters had invalid values.
    """

    if not host:
        host = "localhost"

    if port is not None and port != "":
        port = int(port)
        if not 1 <= port <= 65535:
            raise ConfigError(f"not a valid port number: {repr(port)}")

    if scheme:
        # Scheme specified; if a port is _not_ specified, the we'll try to figure out a
        # default from the scheme.
        scheme = scheme.lower()
        known_ports = KNOWN_SCHEME_PORTS.get(scheme)
        if known_ports is None:
            raise ConfigError(f"not a valid scheme: {repr(scheme)}")
        if port:
            return f"{scheme}://{host}:{port}"
        elif not known_ports:
            raise ConfigError(
                f"there is no default port for {scheme} URLs; you must specify a port"
            )
        else:
            return f"{scheme}://{host}:{known_ports[0]}"

    if port:
        # No scheme provided, but a port was provided.
        # See if we can infer the scheme from the provided port.
        for proposed_scheme, ports in KNOWN_SCHEME_PORTS.items():
            if port in ports:
                return f"{proposed_scheme}://{host}:{port}"

        warnings.warn(
            f"A port ({port}) was specified without supplying a scheme (http/https); "
            "will attempt connecting with https. Unless you are using a standard port, "
            "you should specify a scheme explicitly.",
            ConfigWarning,
        )

    # Neither scheme nor port are provided. If the host is localhost, then assume we're targeting
    # a sandbox; otherwise prefer an SSL/TLS connection.
    return f"grpc://{host}:6865" if is_localhost(host) else f"https://{host}:443"


def is_localhost(host: str) -> bool:
    if host == "localhost":
        return True
    try:
        addr = ipaddress.ip_address(host)
        return addr.is_loopback
    except ValueError:
        return False



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import (
    package_service_pb2 as com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2,
)


class PackageServiceStub(object):
    """Allows clients to query the Daml-LF packages that are supported by the server."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ListPackages = channel.unary_unary(
            "/com.daml.ledger.api.v1.PackageService/ListPackages",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesRequest.SerializeToString,
            response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesResponse.FromString,
        )
        self.GetPackage = channel.unary_unary(
            "/com.daml.ledger.api.v1.PackageService/GetPackage",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageRequest.SerializeToString,
            response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageResponse.FromString,
        )
        self.GetPackageStatus = channel.unary_unary(
            "/com.daml.ledger.api.v1.PackageService/GetPackageStatus",
            request_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusRequest.SerializeToString,
            response_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusResponse.FromString,
        )


class PackageServiceServicer(object):
    """Allows clients to query the Daml-LF packages that are supported by the server."""

    def ListPackages(self, request, context):
        """Returns the identifiers of all supported packages.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the request does not include a valid ledger id
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetPackage(self, request, context):
        """Returns the contents of a single package.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the requested package is unknown
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def GetPackageStatus(self, request, context):
        """Returns the status of a single package.
        Errors:
        - ``UNAUTHENTICATED``: if the request does not include a valid access token
        - ``PERMISSION_DENIED``: if the claims in the token are insufficient to perform a given operation
        - ``NOT_FOUND``: if the requested package is unknown
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_PackageServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "ListPackages": grpc.unary_unary_rpc_method_handler(
            servicer.ListPackages,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesRequest.FromString,
            response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesResponse.SerializeToString,
        ),
        "GetPackage": grpc.unary_unary_rpc_method_handler(
            servicer.GetPackage,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageRequest.FromString,
            response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageResponse.SerializeToString,
        ),
        "GetPackageStatus": grpc.unary_unary_rpc_method_handler(
            servicer.GetPackageStatus,
            request_deserializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusRequest.FromString,
            response_serializer=com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "com.daml.ledger.api.v1.PackageService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class PackageService(object):
    """Allows clients to query the Daml-LF packages that are supported by the server."""

    @staticmethod
    def ListPackages(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.PackageService/ListPackages",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.ListPackagesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetPackage(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.PackageService/GetPackage",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def GetPackageStatus(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/com.daml.ledger.api.v1.PackageService/GetPackageStatus",
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusRequest.SerializeToString,
            com_dot_daml_dot_ledger_dot_api_dot_v1_dot_package__service__pb2.GetPackageStatusResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    Any as __Any,
    Dict as __Map,
    List as _List,
    Literal as __Literal,
    NoReturn as __Void,
    Optional as __Optional,
    Sequence as __Sequence,
    Tuple as __Tuple,
    TypeVar as __TypeVar,
)

from google.protobuf.descriptor import FieldDescriptor as __FieldDescriptor
from google.protobuf.message import Message as __Message

from ........prim import Party

__all__ = [
    "GetParticipantIdRequest",
    "GetParticipantIdResponse",
    "GetPartiesRequest",
    "GetPartiesResponse",
    "ListKnownPartiesRequest",
    "ListKnownPartiesResponse",
    "AllocatePartyRequest",
    "AllocatePartyResponse",
    "PartyDetails",
]

T = __TypeVar("T")

class __List(_List[T]):
    def add(self) -> T: ...

class GetParticipantIdRequest(__Message):
    def __init__(self) -> None: ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetParticipantIdRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Void) -> __Void: ...
    def ClearField(self, field_name: __Void) -> __Void: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetParticipantIdResponse(__Message):
    participant_id: str
    def __init__(self, *, participant_id: __Optional[str] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetParticipantIdResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["participant_id"]) -> bool: ...
    def ClearField(self, field_name: __Literal["participant_id"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetPartiesRequest(__Message):
    parties: __List[Party]
    def __init__(self, *, parties: __Optional[__Sequence[Party]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetPartiesRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["parties"]) -> bool: ...
    def ClearField(self, field_name: __Literal["parties"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class GetPartiesResponse(__Message):
    party_details: __List[PartyDetails]
    def __init__(self, *, party_details: __Optional[__Sequence[PartyDetails]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: GetPartiesResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["party_details"]) -> bool: ...
    def ClearField(self, field_name: __Literal["party_details"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ListKnownPartiesRequest(__Message):
    def __init__(self) -> None: ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ListKnownPartiesRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Void) -> __Void: ...
    def ClearField(self, field_name: __Void) -> __Void: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class ListKnownPartiesResponse(__Message):
    party_details: __List[PartyDetails]
    def __init__(self, *, party_details: __Optional[__Sequence[PartyDetails]] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: ListKnownPartiesResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["party_details"]) -> bool: ...
    def ClearField(self, field_name: __Literal["party_details"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class AllocatePartyRequest(__Message):
    party_id_hint: Party
    display_name: str
    def __init__(self, *, party_id_hint: __Optional[Party] = ..., display_name: __Optional[str] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: AllocatePartyRequest) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["party_id_hint", "display_name"]) -> bool: ...
    def ClearField(self, field_name: __Literal["party_id_hint", "display_name"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class AllocatePartyResponse(__Message):
    party_details: PartyDetails
    def __init__(self, *, party_details: __Optional[PartyDetails] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: AllocatePartyResponse) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["party_details"]) -> bool: ...
    def ClearField(self, field_name: __Literal["party_details"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...

class PartyDetails(__Message):
    party: Party
    display_name: str
    is_local: bool
    def __init__(self, *, party: __Optional[Party] = ..., display_name: __Optional[str] = ..., is_local: __Optional[bool] = ...): ...
    def __eq__(self, other_msg: __Optional[__Any]) -> bool: ...
    def __str__(self) -> str: ...
    def __unicode__(self) -> str: ...
    def MergeFrom(self, other_msg: PartyDetails) -> None: ...
    def Clear(self) -> None: ...
    def SetInParent(self) -> None: ...
    def IsInitialized(self) -> bool: ...
    def MergeFromString(self, serialized: bytes) -> int: ...
    def SerializeToString(self, *, deterministic: bool = ...) -> bytes: ...
    def SerializePartialToString(self, *, deterministic: bool = ...) -> bytes: ...
    def ListFields(self) -> __Sequence[__Tuple[__FieldDescriptor, __Any]]: ...
    def HasField(self, field_name: __Literal["party", "display_name", "is_local"]) -> bool: ...
    def ClearField(self, field_name: __Literal["party", "display_name", "is_local"]) -> None: ...
    def WhichOneof(self, oneof_group: __Void) -> __Void: ...
    def HasExtension(self, extension_handle: __Any) -> bool: ...
    def ClearExtension(self, extension_handle: __Any) -> None: ...
    def UnknownFields(self) -> __Any: ...
    def DiscardUnknownFields(self) -> None: ...
    def ByteSize(self) -> int: ...
    def _SetListener(self, message_listener: __Any) -> None: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from typing import List, Tuple, Union, cast
from urllib.parse import urlparse

from grpc import (
    AuthMetadataContext,
    AuthMetadataPlugin,
    AuthMetadataPluginCallback,
    ChannelCredentials,
    composite_channel_credentials,
    metadata_call_credentials,
    ssl_channel_credentials,
)
from grpc.aio import Channel, insecure_channel, secure_channel

from ..config import Config

__all__ = ["create_channel"]


def create_channel(config: "Config") -> "Channel":
    """
    Create a :class:`Channel` for the specified configuration.
    """
    u = urlparse(config.url.url)

    options = [
        ("grpc.max_send_message_length", -1),
        ("grpc.max_receive_message_length", -1),
    ]
    if not config.url.use_http_proxy:
        options.append(("grpc.enable_http_proxy", 0))

    if (u.scheme in ("https", "grpcs")) or config.ssl:
        credentials = ssl_channel_credentials(
            root_certificates=config.ssl.ca,
            private_key=config.ssl.cert_key,
            certificate_chain=config.ssl.cert,
        )
        if config.access.token:
            # The grpc Credential objects do not actually define a formal interface, and are
            # used interchangeably in the code.
            #
            # Additionally there are some incorrect rules in the grpc-stubs typing rules that force
            # us to work around the type system.
            credentials = cast(
                ChannelCredentials,
                composite_channel_credentials(
                    credentials, metadata_call_credentials(GrpcAuth(config))
                ),
            )
        return secure_channel(u.netloc, credentials, options)
    else:
        return insecure_channel(u.netloc, options)


class GrpcAuth(AuthMetadataPlugin):
    def __init__(self, config: "Config"):
        self._config = config

    def __call__(self, context: "AuthMetadataContext", callback: "AuthMetadataPluginCallback"):
        # This overly verbose type signature is here to satisfy mypy and grpc-stubs
        options = []  # type: List[Tuple[str, Union[str, bytes]]]

        # TODO: Add support here for refresh tokens
        token = self._config.access.token
        if token:
            options.append(("Authorization", "Bearer " + self._config.access.token))

        callback(tuple(options), None)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# fmt: off

from typing import (
    AsyncIterable as __AsyncStream,
    Awaitable as __Awaitable,
    Iterable as __Stream,
    Union as __Union,
    overload,
)

from grpc import Channel as __Channel
from grpc.aio import Channel as __AsyncChannel

from .package_management_service_pb2 import (
    ListKnownPackagesRequest,
    ListKnownPackagesResponse,
    UploadDarFileRequest,
    UploadDarFileResponse,
)

__all__ = [
    "PackageManagementServiceStub",
]

class PackageManagementServiceStub:
    @classmethod
    @overload
    def __new__(cls, channel: __Channel) -> _PackageManagementServiceStub: ...
    @classmethod
    @overload
    def __new__(cls, channel: __AsyncChannel) -> _PackageManagementServiceStub_Async: ...
    def ListKnownPackages(self, __1: ListKnownPackagesRequest) -> __Union[ListKnownPackagesResponse, __Awaitable[ListKnownPackagesResponse]]: ...
    def UploadDarFile(self, __1: UploadDarFileRequest) -> __Union[UploadDarFileResponse, __Awaitable[UploadDarFileResponse]]: ...

class _PackageManagementServiceStub(PackageManagementServiceStub):
    def ListKnownPackages(self, __1: ListKnownPackagesRequest) -> ListKnownPackagesResponse: ...
    def UploadDarFile(self, __1: UploadDarFileRequest) -> UploadDarFileResponse: ...

class _PackageManagementServiceStub_Async(PackageManagementServiceStub):
    def ListKnownPackages(self, __1: ListKnownPackagesRequest) -> __Awaitable[ListKnownPackagesResponse]: ...
    def UploadDarFile(self, __1: UploadDarFileRequest) -> __Awaitable[UploadDarFileResponse]: ...



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

"""
This module contains the Python API for interacting with the Ledger API.
"""
from ast import literal_eval
from configparser import ConfigParser
from pathlib import Path

import pkg_resources

__all__ = [
    "AIOPartyClient",
    "Command",
    "connect",
    "ContractData",
    "ContractId",
    "CreateAndExerciseCommand",
    "CreateCommand",
    "DazlError",
    "ExerciseByKeyCommand",
    "ExerciseCommand",
    "LOG",
    "Network",
    "Party",
    "SimplePartyClient",
    "__version__",
    "async_network",
    "create",
    "create_and_exercise",
    "exercise",
    "exercise_by_key",
    "frozendict",
    "run",
    "setup_default_logger",
    "simple_client",
    "write_acs",
]

from ._logging import LOG
from .client import AIOPartyClient, Network, SimplePartyClient, async_network, run, simple_client
from .client.commands import (
    CreateAndExerciseCommand,
    CreateCommand,
    ExerciseByKeyCommand,
    ExerciseCommand,
    create,
    create_and_exercise,
    exercise,
    exercise_by_key,
)
from .ledger import Command
from .ledger.grpc import connect
from .model.core import ContractId
from .pretty.table import write_acs
from .prim import ContractData, DazlError, FrozenDict as frozendict, Party
from .util.logging import setup_default_logger

try:
    # This method is undocumented, but is required to read large size of model files when using
    # the C++ implementation.
    # noinspection PyPackageRequirements,PyUnresolvedReferences,PyProtectedMember
    from google.protobuf.pyext import _message

    _message.SetAllowOversizeProtos(True)

except ImportError:
    # ImportError for the Protobuf libraries is likely fatal, but this would not be the most helpful
    # place to throw an ImportError.
    pass


def _get_version() -> str:
    """
    Used to make the version of this library easily accessible programmatically.
    Two techniques are tried:
     1. Try to read it from the current package definition. This is what is used
        when trying to look up version information if dazl is installed via a
        wheel file.
     2. Use the value from the local pyproject.toml file (this is used when
        running dazl from source).
    """
    try:
        return pkg_resources.require("dazl")[0].version
    except pkg_resources.DistributionNotFound:
        pass
    except Exception:
        return "unknown"

    try:
        config = ConfigParser()
        config.read(Path(__file__).parent.parent / "pyproject.toml")
        if "tool.poetry" in config:
            poetry_section = config["tool.poetry"]
            return literal_eval(poetry_section["version"])
    except Exception:  # noqa
        pass

    return "unknown"


__version__ = _get_version()



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/daml/ledger/api/v1/admin/participant_pruning_service.proto

from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    symbol_database as _symbol_database,
)

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="com/daml/ledger/api/v1/admin/participant_pruning_service.proto",
    package="com.daml.ledger.api.v1.admin",
    syntax="proto3",
    serialized_options=b"\n\034com.daml.ledger.api.v1.adminB#ParticipantPruningServiceOuterClass\252\002\034Com.Daml.Ledger.Api.V1.Admin",
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n>com/daml/ledger/api/v1/admin/participant_pruning_service.proto\x12\x1c\x63om.daml.ledger.api.v1.admin":\n\x0cPruneRequest\x12\x13\n\x0bprune_up_to\x18\x01 \x01(\t\x12\x15\n\rsubmission_id\x18\x02 \x01(\t"\x0f\n\rPruneResponse2}\n\x19ParticipantPruningService\x12`\n\x05Prune\x12*.com.daml.ledger.api.v1.admin.PruneRequest\x1a+.com.daml.ledger.api.v1.admin.PruneResponseBb\n\x1c\x63om.daml.ledger.api.v1.adminB#ParticipantPruningServiceOuterClass\xaa\x02\x1c\x43om.Daml.Ledger.Api.V1.Adminb\x06proto3',
)


_PRUNEREQUEST = _descriptor.Descriptor(
    name="PruneRequest",
    full_name="com.daml.ledger.api.v1.admin.PruneRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="prune_up_to",
            full_name="com.daml.ledger.api.v1.admin.PruneRequest.prune_up_to",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="submission_id",
            full_name="com.daml.ledger.api.v1.admin.PruneRequest.submission_id",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=96,
    serialized_end=154,
)


_PRUNERESPONSE = _descriptor.Descriptor(
    name="PruneResponse",
    full_name="com.daml.ledger.api.v1.admin.PruneResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=156,
    serialized_end=171,
)

DESCRIPTOR.message_types_by_name["PruneRequest"] = _PRUNEREQUEST
DESCRIPTOR.message_types_by_name["PruneResponse"] = _PRUNERESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

PruneRequest = _reflection.GeneratedProtocolMessageType(
    "PruneRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _PRUNEREQUEST,
        "__module__": "com.daml.ledger.api.v1.admin.participant_pruning_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.PruneRequest)
    },
)
_sym_db.RegisterMessage(PruneRequest)

PruneResponse = _reflection.GeneratedProtocolMessageType(
    "PruneResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _PRUNERESPONSE,
        "__module__": "com.daml.ledger.api.v1.admin.participant_pruning_service_pb2"
        # @@protoc_insertion_point(class_scope:com.daml.ledger.api.v1.admin.PruneResponse)
    },
)
_sym_db.RegisterMessage(PruneResponse)


DESCRIPTOR._options = None

_PARTICIPANTPRUNINGSERVICE = _descriptor.ServiceDescriptor(
    name="ParticipantPruningService",
    full_name="com.daml.ledger.api.v1.admin.ParticipantPruningService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=173,
    serialized_end=298,
    methods=[
        _descriptor.MethodDescriptor(
            name="Prune",
            full_name="com.daml.ledger.api.v1.admin.ParticipantPruningService.Prune",
            index=0,
            containing_service=None,
            input_type=_PRUNEREQUEST,
            output_type=_PRUNERESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_PARTICIPANTPRUNINGSERVICE)

DESCRIPTOR.services_by_name["ParticipantPruningService"] = _PARTICIPANTPRUNINGSERVICE

# @@protoc_insertion_point(module_scope)



# Copyright (c) 2017-2021 Digital Asset (Switzerland) GmbH and/or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from types import MappingProxyType
from typing import Collection, Mapping, Optional
import warnings

from .daml_lf_1 import Expr, Type, ValName

with warnings.catch_warnings():
    warnings.simplefilter("ignore", DeprecationWarning)
    from ..model.types_store import PackageStore


warnings.warn(
    "dazl.damlast.eval_scope is deprecated; there is no replacement.",
    DeprecationWarning,
    stacklevel=2,
)


class EvaluationScope:
    def __init__(
        self,
        store: "PackageStore",
        bindings: "Mapping[str, Expr]",
        blocked_value_refs: "Collection[ValName]" = (),
        depth: int = 0,
    ):
        warnings.warn(
            "EvaluationScope is deprecated; there is no replacement.",
            DeprecationWarning,
            stacklevel=2,
        )
        self.store = store
        self.depth = depth
        self._vars = MappingProxyType(dict(bindings))
        self._blocked_value_refs = frozenset(blocked_value_refs)

    def without(self, var_name: str) -> "EvaluationScope":
        """
        Return a child scope with the specified variable name _unbound_.

        :param var_name:
        :return:
        """
        new_bindings = dict(self._vars)
        del new_bindings[var_name]
        return EvaluationScope(self.store, new_bindings, self._blocked_value_refs, self.depth + 1)

    def without_value(self, value_ref) -> "EvaluationScope":
        """
        Return a child scope where the value corresponding to the specified value reference will
        NOT be returned. (This is used to handle breaking otherwise recursive calls.)

        :param value_ref: The value to withhold.
        """
        values = set(self._blocked_value_refs)
        values.discard(value_ref)

        return EvaluationScope(self.store, self._vars, frozenset(values), self.depth + 1)

    def child_scope(self, new_bindings: "Mapping[str, Expr]") -> "EvaluationScope":
        bindings = dict(self._vars)
        bindings.update(new_bindings)
        return EvaluationScope(self.store, bindings, self._blocked_value_refs, self.depth + 1)

    def get_bound(self, var_name) -> "Expr":
        return self._vars[var_name]

    def resolve_type(self, type_ref) -> "Type":
        pass

    def resolve_value(self, value_ref) -> "Optional[Expr]":
        if value_ref in self._blocked_value_refs:
            return None
        else:
            return self.store.resolve_value_reference(value_ref)
